from fieldday import QTY_CAPABLE
if QTY_CAPABLE:
    from fieldday import ureg, Q_  # from pint via __init__.py


class Field:
    def __init__(self, value, desc='', units='', fmt='', **otherkwargs: object):
        """ Base class for a data item.
        This class and its subclasses are meant for maniupulating fields passed around an IoT system, e.g. MQTT as
        JSON payloads.  They provide the means for:
        - decoding from numeric or string values via their constructor,
        - encoding to numeric or string values via an encode() method, and
        - formatting output in various ways
        This class and its subclasses should always be able to construct (decode) from a string representation.
        In cases where the field's magnitude can be represented as a more compact numeric value, the encoder
        should generate a numeric value and construct (decode) from a numeric value as well as a string.
        This class is meant to be subclassed.  For example, you can add a convert() method to convert
        a class to different units ot to a different type of field class.
        Besides `value`, `desc`, and `units`, other attributes can be user-defined with `otherkwargs`.
        Args:
            value (obj): stored data value (can be any object)
            desc (str): a short description of the field
            units (str): measurement units (if any)
            fmt (str): format string for output of __str__ magnitude component only
            otherkwargs (dict): dynamically user-defined application-specific attributes
        """
        self.value = value
        self.desc = desc
        self.units = units
        self.fmt = fmt
        self.__dict__.update(otherkwargs) # anything else

    def encode(self):
        """
        Returns: The encoded field as a JSON-encodable object and the most compact object that that can be decoded
        by the constructor
        """
        return self.value

    def __str__(self):
        """
        Returns: the encoded value as a string with a specific format if provided
        """
        s = ''
        if self.desc:
            s = s + self.desc + ': '
        s = s + str(self.str_value())
        if self.units:
            s = s + ' ' + self.units
        return s

    def str_value(self):
        s = '{0'
        if self.fmt:
            s = s + ':' + self.fmt
        s = s + '}'
        return s.format(self.value)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.encode()!r}, {self.desc!r}, {self.units!r}, {self.fmt!r})"


class FieldFloat(Field):
    """ Basic numeric (`float`) field """
    def __init__(self, encoded_value, **kwargs):
        super().__init__(float(encoded_value), **kwargs)


class FieldInt(Field):
    """ Basic `int` field
    optional keyword arguments:
        base (int): if encoded_value is a string, this can represent the base as second parameter of int()
    """
    def __init__(self, encoded_value, **kwargs):
        if isinstance(encoded_value, str):
            base = kwargs.pop('base', 0)
            value = int(encoded_value, base)
        else:
            value = int(encoded_value)
        super().__init__(value, **kwargs)


class FieldStr(Field):
    """ Basic numeric (`float`) field """
    def __init__(self, encoded_value, **kwargs):
        super().__init__(str(encoded_value), **kwargs)



class FieldQty(Field):
    """ Field in which the units are tied to the encoded value.
    The units attribute of instances of this class will always be empty.  The advantage of using this
    field type is that the units can be changed upstream and are not locked into the field type metadata.
    Uses a `pint.UnitRegistry.Quantity` as the underlying data store so we can do unit conversions easily """
    def __init__(self, encoded_value, **kwargs):
        if not QTY_CAPABLE:
            raise Exception("Not Qty-enabled.  You probably need to install Pint.")
        if isinstance(encoded_value, FieldFloat) or isinstance(encoded_value, FieldInt):
            super().__init__(Q_(str(encoded_value.value)+" "+str(encoded_value.units)),
                             desc=encoded_value.desc, units='', fmt=encoded_value.fmt)
        else:
            super().__init__(Q_(encoded_value), **kwargs)
        if 'units' in kwargs:
            raise ValueError("Units are tied to FieldQty values so the desc attribute should not be used")

    def encode(self):
        return str(self.value)

    def str_value(self):
        return ('{0:' + self.fmt + '} {1}').format(self.value.magnitude, str(self.value.units))

    def str_mag(self):
        return str(self.value.magnitude)

    def str_units(self):
        return str(self.value.units)

    def convert(self, new_units, new_fmt=None):
        """ Convert units using `pint` to intelligently transform to a new units string """
        self.value.ito(ureg.parse_units(new_units))
        if new_fmt is not None:
            self.fmt = new_fmt






try:
    from pint import UnitRegistry
    QTY_CAPABLE = True
    # instantiate the pint registry only once
    ureg = UnitRegistry()
    Q_ = ureg.Quantity
    # import the other internal modules afterwords because they import the instances ureg and Q_
except ImportError as e:
    QTY_CAPABLE = False

from fieldday.fieldclasses import *
from fieldday.helpers import *
from fieldday.fieldset import FieldSet



import json
import fieldday.helpers as helpers


class FieldSet(dict):
    """
    A set (dict) of fields.  If field_types is provided, this class will store the set as
    a dict of field types.
    The built-in methods can provide the user with two versions of the field values:
        1. Encoded: This is the encoded JSON version of the field set, for transmission. Accessed with encode().
        2. Decoded: This is the human-readable version.  Accessed with the str() or print() methods.
    Examples:

    """

    def __init__(self, input_field_set, rename_dict=None, field_types=None):
        """
        Instantiate a FieldSet from an encoded list.
        Nominally, decode from a JSON string, but also can work from a `dict`.
        Args:
            input_field_set (str, dict): set of input fields
            rename_dict (dict): dict of field names to rename (key is original field name)
            field_types (dict): set of tuples (field types, kwargs) (key is renamed field name)
        """
        if isinstance(input_field_set, str):
            # assume it's a JSON object (an unordered collection of name–value pairs)
            field_dict = json.loads(input_field_set)
        else:
            if isinstance(input_field_set, dict):
                field_dict = input_field_set.copy()
            else:
                raise ValueError('input is not a JSON object or dict')
        if rename_dict is not None:
            helpers.rename_keys(field_dict, rename_dict)
        if field_types is not None:
            field_dict = self.load_encoded_dict(field_dict, field_types)
        super().__init__(field_dict)

    def encode(self):
        """
        Returns: self as a json string of encoded values - can then json.load() into a dict (key is field name)
        """
        return json.dumps(self, default=lambda f: f.encode())

    @staticmethod
    def load_encoded_dict(raw_dict, fieldtypes):
        """ load an encoded dictionary and return a dictionary of Fields """
        newdict = {}
        for key, val in raw_dict.items():
            if key in fieldtypes:
                fielddef = fieldtypes[key]
                newdict[key] = fielddef[0](val, **fielddef[1])
            else:
                newdict[key] = val
        return newdict

    def modify_fields(self, field_modifier_dict):
        """ modify specific fields in this set in place with modifier functions
        Args:
            field_modifier_dict (dict): a dict of modifier functions (key is field names to modify)
        """
        for key, modifier in field_modifier_dict.items():
            if key in self:
                field_modifier_dict[key](self[key])

    def modify_types(self, types_modifier_list):
        """ modify specific field types in this set in place with modifier functions if a criteria is met
        Args:
            types_modifier_list (list): a list of lists or tuples (field_type, modifier, criteria_checker)
                field_type: the type of field to modify
                modifier: a reference to the method that will perform the modification
                criteria_checker: a reference to the method that must return True for the modification to happen
        """
        for item in types_modifier_list:
            for key, val in self.items():
                if isinstance(val, item[0]):
                    if len(item) > 2:
                        meetscriteria = item[2](self[key])
                    else:
                        meetscriteria = True
                    if meetscriteria:
                        item[1](self[key])

    def cast_fields(self, new_field_types):
        """ modify specific fields in this set in place with modifier functions
        Args:
            new_field_types (dict): a dict of new field types (key is field names to modify)
        """
        for key, modifier in new_field_types.items():
            if key in self:
                self[key] = new_field_types[key](self[key])

    def __str__(self):
        # use .join() to avoid a newline at the end of the list
        slist = []
        for key in self:
            slist.append(key + ': ' + str(self[key]))
        return '\n'.join(slist)

    # def replace_fields(self, field_convert_dict):
    #     """ Converty """
    #     pass
    #
    # def replace_types(self, field_types_dict):
    #     """ Modify a field set using a list of types """
    #     pass

    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



def rename_keys(field_dict: dict, key_replacement_map: dict):
    """ Replaces keys in a dictionary according to a key replacement map """
    for key in list(field_dict.keys()):
        if key in key_replacement_map:
            field_dict[key_replacement_map[key]] = field_dict.pop(key)


def encode_field(field):
    """ for JSON.dumps """
    return field.encode()




import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# with open('requirements.txt', 'r') as fin:
#     reqs = fin.read().splitlines()
# print(f"requirements: {reqs}")

# This call to setup() does all the work
setup(
    name="fieldday",
    version="0.0.1",
    description="Data field manager for IoT applications",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/jmfife/fieldday",
    author="Mike Fife",
    author_email="jmfife@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["fieldday"],
    include_package_data=True,
    entry_points={
    },
    install_requires=[],
    python_requires='>=3.7',
    extras_require={
        "qty": ["pint"],
    }
)